import uuid

from django.contrib import messages
from django.core.cache import cache
from django.db.models import Q
from django.http import HttpResponseRedirect
from django.utils.translation import ugettext_lazy as _
from django.views.generic import TemplateView

from .forms import SearchAndReplaceForm


class SearchAndReplaceView(TemplateView):
    template_name = "search_and_replace/search_and_replace.html"
    form_class = SearchAndReplaceForm
    models_and_fields = None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["title"] = _("Search and replace")
        context["form"] = self.form
        return context

    def get_form(self, request):
        return self.form_class(
            self.models_and_fields, data=request.POST if request.POST else None
        )

    def get(self, request, *args, **kwargs):
        self.form = self.get_form(request)
        return super().get(request, *args, **kwargs)

    def is_double_submit(self):
        preview_id = self.request.POST.get("preview_id")
        if not preview_id:
            return True

        cache_key = "search-replace-{}".format(preview_id)
        if cache.get(cache_key):
            return True
        else:
            cache.set(cache_key, "done")
            return False

    def post(self, request, *args, **kwargs):
        self.form = self.get_form(request)
        if self.form.is_valid():
            if "apply" in request.POST and not self.is_double_submit():
                return self.form_valid(preview=False)
            else:
                return self.form_valid(preview=True)
        return self.render_to_response(self.get_context_data())

    def get_query_set(self, model):
        return model._default_manager.all()

    def filter_qs(self, search, model, fields):
        assert fields
        qs = self.get_query_set(model)
        filter = Q()
        for field in fields:
            filter |= Q(**{"{}__contains".format(field): search})
        return qs.filter(filter)

    def apply_search_and_replace(self, search, replace, model, fields, preview):
        qs = self.filter_qs(search, model, fields)
        results = []
        for instance in qs:
            changed_fields = []
            for field in fields:
                value = getattr(instance, field)

                # support for markup fields
                if hasattr(value, "raw"):
                    value = value.raw

                if search in value:
                    new_value = value.replace(search, replace)
                    setattr(instance, field, new_value)
                    changed_fields.append((field, value, new_value))
            results.append((instance, changed_fields))
            if not preview:
                instance.save()
        return results

    def get_results(self, search, replace, preview=True):
        results = []
        for model, selected_fields in self.form.get_selected_fields():
            result = self.apply_search_and_replace(
                search, replace, model, selected_fields, preview=preview
            )
            if result:
                results.append((model, result))
        return results

    def form_valid(self, preview=True):
        search = self.form.cleaned_data["search"]
        replace = self.form.cleaned_data["replace"]

        results = self.get_results(search, replace, preview=preview)
        num_results = sum(len(instances) for model, instances in results)

        if preview:
            return self.response_preview(search, replace, results, num_results)
        else:
            return self.response_success(search, replace, results, num_results)

    def response_preview(self, search, replace, results, num_results):
        return self.render_to_response(
            self.get_context_data(
                results=results,
                num_results=num_results,
                search=search,
                replace=replace,
                preview_id=str(uuid.uuid4()),
            )
        )

    def response_success(self, search, replace, results, num_results):
        messages.success(
            self.request,
            _("Replaced {} with {} in {} instances").format(
                search, replace, num_results
            ),
        )
        return HttpResponseRedirect(".")



from django import forms
from django.db.models.options import Options


class SearchAndReplaceForm(forms.Form):
    search = forms.CharField()
    replace = forms.CharField(required=False)

    def _get_form_field_name(self, model, field):
        opts = model._meta
        assert isinstance(opts, Options)
        return "{}_{}_{}".format(opts.app_label, opts.model_name, field)

    def _get_field_label(self, model, field):
        opts = model._meta
        assert isinstance(opts, Options)
        return opts.get_field(field).verbose_name

    def get_form_fields_by_model(self):
        """
        Returns a list of (model verbose name, [form field, ...]) tuples.
        Used to render the form.
        """
        results = []
        for model, fields in self.models_and_fields:
            form_fields = []
            for field in fields:
                form_field_name = self._get_form_field_name(model, field)
                form_fields.append(self[form_field_name])
            results.append((model, form_fields))
        return results

    def get_extra_form_fields(self):
        """
        Returns all fields except search, replace and the ones returned by get_form_fields_by_model.
        """
        ignored = {"search", "replace"}.union(
            {
                self._get_form_field_name(model, field)
                for model, fields in self.models_and_fields
                for field in fields
            }
        )
        fields = []
        for field in self.fields:
            if field not in ignored:
                fields.append(self[field])
        return fields

    def get_selected_fields(self):
        """
        Returns a list of (Model, [field name, ...]) tuples that were selected by the user.
        Used by the SearchAndReplaceView to decide which models / fields to target.
        """
        results = []
        for model, fields in self.models_and_fields:
            selected = []
            for field in fields:
                form_field_name = self._get_form_field_name(model, field)
                if self.cleaned_data[form_field_name]:
                    selected.append(field)
            if selected:
                results.append((model, selected))
        return results

    def __init__(self, models_and_fields, data=None, files=None, **kwargs):
        self.models_and_fields = models_and_fields
        super().__init__(data, files, **kwargs)
        for model, fields in models_and_fields:
            for field in fields:
                form_field_name = self._get_form_field_name(model, field)
                field_label = self._get_field_label(model, field)
                self.fields[form_field_name] = forms.BooleanField(
                    required=False, label=field_label, initial=True
                )



from django import forms
from django.contrib import admin
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.urls import path
from django.db import models
from django.test import TestCase, RequestFactory

from search_and_replace.forms import SearchAndReplaceForm
from search_and_replace.views import SearchAndReplaceView


urlpatterns = [path("admin/", admin.site.urls)]  # noqa, used for runtests.py


class Cat(models.Model):
    name = models.CharField(max_length=255)
    bio = models.TextField(verbose_name="Biography")

    def __str__(self):
        return self.name


class Dog(models.Model):
    name = models.CharField(max_length=255)
    bark = models.TextField(verbose_name="Bark")

    def __str__(self):
        return self.name


class SearchAndReplaceFormTest(TestCase):
    def setUp(self):
        self.models_and_fields = [(Cat, ("name", "bio")), (Dog, ("name", "bark"))]
        self.form = SearchAndReplaceForm(self.models_and_fields)

    def test_form_field_name(self):
        self.assertEqual(
            self.form._get_form_field_name(Cat, "name"), "search_and_replace_cat_name"
        )

    def test_form_field_label(self):
        self.assertEqual(self.form._get_field_label(Cat, "bio"), "Biography")

    def test_get_form_fields_by_model(self):
        (cat, cat_fields), (dog, dog_fields) = self.form.get_form_fields_by_model()
        self.assertEqual(cat, Cat)
        self.assertEqual(dog, Dog)

        self.assertEqual(cat_fields[0].name, "search_and_replace_cat_name")
        self.assertEqual(cat_fields[1].name, "search_and_replace_cat_bio")

        self.assertEqual(dog_fields[0].name, "search_and_replace_dog_name")
        self.assertEqual(dog_fields[1].name, "search_and_replace_dog_bark")

    def test_get_extra_form_fields_is_empty_by_default(self):
        self.assertEqual(self.form.get_extra_form_fields(), [])

    def test_get_extra_form_fields_has_additional_fields(self):
        class ExtendedForm(SearchAndReplaceForm):
            extra_1 = forms.CharField()
            extra_2 = forms.CharField()

        first, second = ExtendedForm(self.models_and_fields).get_extra_form_fields()
        self.assertEqual(first.name, "extra_1")
        self.assertEqual(second.name, "extra_2")

    def test_get_selected_fields(self):
        form = SearchAndReplaceForm(
            self.models_and_fields,
            data={
                "search": "term",
                "search_and_replace_dog_bark": True,
                "search_and_replace_dog_name": False,
                "search_and_replace_cat_name": True,
                "wrong_name": True,
            },
        )
        assert form.is_valid()

        selected = form.get_selected_fields()
        self.assertEqual(selected, [(Cat, ["name"]), (Dog, ["bark"])])


class SearchAndReplaceViewTest(TestCase):
    def setUp(self):
        self.models_and_fields = [(Cat, ("name", "bio")), (Dog, ("name", "bark"))]

        self.lucy = Cat.objects.create(name="Lucy", bio="Grew up in the deep south.")
        self.peter = Cat.objects.create(name="Peter", bio="Child of the north.")
        self.momo = Cat.objects.create(name="Momo the first.", bio="You know me.")
        self.george = Dog.objects.create(name="George", bark="Whoof!")
        self.adam = Dog.objects.create(name="Adam", bark="Whef, the whef!")

    def test_render_search_and_replace_form(self):
        request = RequestFactory().get("/")

        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, "search_and_replace_dog_name")
        self.assertContains(response, "search_and_replace_dog_bark")

    def test_search_and_replace_form_without_preview_does_not_contain_preview_id(self):
        request = RequestFactory().get("/")

        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertNotContains(response, "preview_id")

    def test_search_and_replace_after_submit_contains_preview_id(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "true",
            },
        )
        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, 'name="preview_id" value="')

    def test_render_search_and_replace_form_preview_contains_the_right_instances(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "true",
            },
        )

        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, "Lucy")
        self.assertContains(response, "Peter")
        self.assertContains(response, "Momo")
        self.assertContains(response, "Adam")
        self.assertNotContains(response, "George")

    def test_render_search_and_replace_form_preview_matches_the_right_fields(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "false",
            },
        )

        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, "Lucy")
        self.assertContains(response, "Peter")
        self.assertContains(response, "Adam")
        self.assertNotContains(response, "Momo")
        self.assertNotContains(response, "George")

    def test_render_search_and_replace_form_preview_contains_the_right_values(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "replace": "teh",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "true",
            },
        )

        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, "the deep south")
        self.assertContains(response, "teh deep south")
        self.assertContains(response, "the north")
        self.assertContains(response, "teh north")
        self.assertContains(response, "the first")
        self.assertContains(response, "teh first")
        self.assertContains(response, "the whef")
        self.assertContains(response, "teh whef")

    def test_preview_does_not_change_values(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "replace": "teh",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "true",
            },
        )

        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)

        self.lucy.refresh_from_db()

        self.assertContains(response, "the deep south")
        self.assertContains(response, "teh deep south")
        self.assertIn("the", self.lucy.bio)
        self.assertNotIn("teh", self.lucy.bio)

    def test_render_invalid_form(self):
        request = RequestFactory().post("/", {})
        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, "form")

    def test_missing_preview_id_replaces_nothing(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "replace": "teh",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "true",
                "apply": "true",
            },
        )
        SessionMiddleware().process_request(request)
        MessageMiddleware().process_request(request)

        SearchAndReplaceView.as_view(models_and_fields=self.models_and_fields)(request)

        self.lucy.refresh_from_db()

        self.assertEqual(self.lucy.bio, "Grew up in the deep south.")

    def test_replace_text_replaces_text(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "the",
                "replace": "teh",
                "search_and_replace_dog_bark": "true",
                "search_and_replace_dog_name": "true",
                "search_and_replace_cat_bio": "true",
                "search_and_replace_cat_name": "true",
                "apply": "true",
                "preview_id": "an-id-used-to-prevent-double-submit",
            },
        )
        SessionMiddleware().process_request(request)
        MessageMiddleware().process_request(request)

        SearchAndReplaceView.as_view(models_and_fields=self.models_and_fields)(request)

        self.lucy.refresh_from_db()
        self.momo.refresh_from_db()
        self.adam.refresh_from_db()

        self.assertEqual(self.lucy.bio, "Grew up in teh deep south.")
        self.assertEqual(self.momo.name, "Momo teh first.")
        self.assertEqual(self.adam.bark, "Whef, teh whef!")

    def test_search_is_case_sensitive(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "South",
                "replace": "east",
                "search_and_replace_cat_bio": "true",
            },
        )
        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertNotContains(response, "Lucy")

        request = RequestFactory().post(
            "/",
            {
                "search": "south",
                "replace": "east",
                "search_and_replace_cat_bio": "true",
            },
        )
        response = SearchAndReplaceView.as_view(
            models_and_fields=self.models_and_fields
        )(request)
        self.assertContains(response, "Lucy")

    def test_prevent_double_submit(self):
        request = RequestFactory().post(
            "/",
            {
                "search": "Lucy",
                "replace": "Lucy Lucy",
                "search_and_replace_cat_name": "true",
                "apply": "true",
                "preview_id": "another-id-used-in-a-double-submit",
            },
        )
        SessionMiddleware().process_request(request)
        MessageMiddleware().process_request(request)
        response = SearchAndReplaceView.as_view(models_and_fields=self.models_and_fields)(request)

        self.assertEqual(response.status_code, 302)

        self.lucy.refresh_from_db()
        self.assertEqual(self.lucy.name, "Lucy Lucy")

        request = RequestFactory().post(
            "/",
            {
                "search": "Lucy",
                "replace": "Lucy Lucy",
                "search_and_replace_cat_name": "true",
                "apply": "true",
                "preview_id": "another-id-used-in-a-double-submit",
            },
        )
        SessionMiddleware().process_request(request)
        MessageMiddleware().process_request(request)
        response = SearchAndReplaceView.as_view(models_and_fields=self.models_and_fields)(request)

        self.assertEqual(response.status_code, 200)

        self.lucy.refresh_from_db()
        self.assertEqual(self.lucy.name, "Lucy Lucy")
